#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  load_config.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
'''
Module containing `LoadConfig` Pydantic model
'''
### Standard Packages ###
from typing import Optional
### Third-Party Packages ###
from pydantic import BaseModel, validator, StrictInt, StrictStr

class LoadConfig(BaseModel):
  backend: Optional[StrictStr] = 'inmemory'
  ttl: Optional[StrictInt]     = 60 # seconds

  ### Redis ###
  redis_url: Optional[StrictStr]

  ### Memcached ###
  memcached_host: Optional[StrictStr]

  ### AWS DynamoDB & MongoDB ###
  table_name: Optional[StrictStr]

  ### AWS DynamoDB ###
  region: Optional[StrictStr]
  dynamodb_url: Optional[StrictStr]

  ### MongoDB ###
  database_name: Optional[StrictStr]
  mongodb_url: Optional[StrictStr]

  @validator('backend')
  def validate_backend(cls, value: str):
    if value.lower() not in { 'dynamodb', 'inmemory', 'memcached', 'redis', 'mongodb' }:
      raise ValueError(
        'The "backend" value must be one of "dynamodb", "inmemory", "memcached", or "redis".'
      )
    return value.lower()

  @validator('ttl')
  def validate_time_to_live(cls, value: str):
    if value <= 0 or value > 3600:
      raise ValueError('The "ttl" value must between 1 or 3600 seconds.')
    return value

  @validator('redis_url', always=True)
  def validate_redis_url(cls, value: str, values: dict):
    if values['backend'].lower() == 'redis' and not value:
      raise ValueError('The "redis_url" cannot be null when using redis as backend.')
    ### TODO More Validations ###
    return value

  @validator('memcached_host', always=True)
  def validate_memcached_host(cls, value: str, values: dict):
    if values['backend'].lower() == 'memcached' and not value:
      raise ValueError('The "memcached_host" cannot be null when using memcached as backend.')
    ### TODO More Validations ###
    return value

  @validator('table_name')
  def validate_table_name(cls, value: str, values: dict):
    backend: str = values['backend'].lower()
    if backend in ('dynamodb', 'mongodb') and not value:
      raise ValueError('The "table_name" cannot be null when using DynamoDB / MongoDB as backend.')
    ### TODO More Validations ###
    return value

  @validator('region')
  def validate_region(cls, value: str, values: dict):
    backend: str      = values['backend'].lower()
    dynamodb_url: str = values.get('dynamodb_url', None)
    if backend == 'dynamodb' and not dynamodb_url and not value:
      raise ValueError(
        'The "region" cannot be null when using DynamoDB as backend and no "dynamodb_url" defined.'
      )
    if value.lower() not in {
      'us-east-1', 'us-west-2', 'us-west-1', 'eu-west-1',   \
        'eu-central-1', 'ap-southeast-1', 'ap-southeast-2', \
          'ap-northeast-1', 'ap-northeast-2', 'sa-east-1',  \
            'cn-north-1', 'ap-south-1'                      \
    }:
      raise ValueError('The "region" provided does not exist under AWS Regions.')
    return value.lower()

  @validator('dynamodb_url', always=True)
  def validate_dynamodb_url(cls, value: str, values: dict):
    backend: str = values['backend'].lower()
    region: str  = values.get('region', None)
    if backend == 'dynamodb' and not region and not value:
      raise ValueError(
        'The "dynamodb_url" cannot be null when using DynamoDB as backend and no region defined.'
      )
    ### TODO More Validations ###
    return value

  @validator('database_name')
  def validate_database_name(cls, value: str, values: dict):
    backend: str = values['backend'].lower()
    if backend == 'mongodb' and not value:
      raise ValueError('The "database_name" cannot be null when using MongoDB as backend.')
    ### TODO More Validations ###
    return value

  @validator('mongodb_url', always=True)
  def validate_mongodb_url(cls, value: str, values: dict):
    backend: str = values['backend'].lower()
    if backend == 'mongodb' and not value:
      raise ValueError('The "mongodb_url" cannot be null when using MongoDB as backend.')
    ### TODO More Validations ###
    return value



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  backends/redis.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
### Standard Packages ###
from dataclasses import dataclass
from typing import Optional, Tuple
### Third-Party Pacakges ###
from redis.asyncio import Redis
### Local Modules ###
from fastapi_cachette.backends import Backend

@dataclass
class RedisBackend(Backend):
  redis: Redis
  ttl: int

  @classmethod
  async def init(cls, redis_url: str, ttl: int) -> 'RedisBackend':
    return cls(Redis.from_url(url=redis_url), ttl)

  async def fetch(self, key) -> str:
    return await self.redis.get(key)

  async def fetch_with_ttl(self, key: str) -> Tuple[int, str]:
    async with self.redis.pipeline(transaction=True) as pipe:
      return await (pipe.ttl(key).get(key).execute())

  async def put(self, key: str, value: str, ttl: Optional[int] = None):
    return await self.redis.set(key, value, ex=(ttl or self.ttl))

  async def clear(self, namespace: Optional[str] = None, key: Optional[str] = None) -> int:
    if namespace:
      lua: str = \
        f'for i, key in ipairs(redis.call("KEYS", "{namespace}:*")) do redis.call("DEL", key); end'
      return await self.redis.eval(lua, numkeys=0)
    elif key:
      return await self.redis.delete(key)



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  backends/memcached.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
### Standard Packages ###
from dataclasses import dataclass
from typing import Optional, Tuple
### Third-Party Packages ###
from aiomcache import Client
### Local Modules ###
from fastapi_cachette.backends import Backend

@dataclass
class MemcachedBackend(Backend):
  mcache: Client
  ttl: int

  @classmethod
  async def init(cls, memcached_host: str, ttl: int) -> 'MemcachedBackend':
    return MemcachedBackend(Client(host=memcached_host), ttl)

  async def fetch(self, key: str):
    return await self.mcache.get(key.encode())

  async def fetch_with_ttl(self, key: str) -> Tuple[int, str]:
    return 3600, await self.mcache.get(key.encode())

  async def put(self, key: str, value: str, ttl: Optional[int] = None):
    return await self.mcache.set(key.encode(), value.encode(), exptime=ttl or self.ttl)

  async def clear(self, namespace: Optional[str] = None, key: Optional[str] = None):
    count: int = 0
    if namespace:
      raise NotImplementedError
    elif key:
      count += (0, 1)[await self.mcache.delete(key.encode())]      
    return count



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  core.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
'''
Module containing Core implementation for Cashette extension for FastAPI
'''
### Standard Packages ###
from asyncio import run
### Third-Party Packages ###
from fastapi.requests import Request
from fastapi.responses import Response
### Local Modules ###
from fastapi_cachette.backends import Backend
from fastapi_cachette.cachette_config import CachetteConfig

class Cachette(CachetteConfig):
  backend: Backend

  def __init__(self, request: Request, response: Response):
    '''
    Invoked by FastAPI Depends
    '''
    # TODO Check request headers if `Cache-Control`` is `no-store`
    if self._backend == 'dynamodb':
      from fastapi_cachette.backends.dynamodb import DynamoDBBackend
      self.backend = run(DynamoDBBackend.init(
        self._table_name, self._ttl, self._region, self._dynamodb_url
      ))
    elif self._backend == 'inmemory':
      from fastapi_cachette.backends.inmemory import InMemoryBackend
      self.backend = InMemoryBackend(self._ttl)
    elif self._backend == 'memcached':
      from fastapi_cachette.backends.memcached import MemcachedBackend
      self.backend = run(MemcachedBackend.init(self._memcached_host, self._ttl))
    elif self._backend == 'mongodb':
      from fastapi_cachette.backends.mongodb import MongoDBBackend
      self.backend = run(MongoDBBackend.init(
        self._database_name, self._table_name, self._ttl, self._mongodb_url
      ))
    elif self._backend == 'redis':
      from fastapi_cachette.backends.redis import RedisBackend
      self.backend = run(RedisBackend.init(self._redis_url, self._ttl))

  # When a method is not found, shortcuts it to instance's `backend` member method
  def __getattr__(self, name):
    return getattr(self.backend, name)



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  backends/__init__.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
'''
Module containing `Backend` abstract class to be inherited by implementation-specific backends
'''
### Standard Packages ###
from abc import abstractmethod
from time import time
from typing import Optional, Tuple

class Backend:

  @property
  def now(self) -> int:
    return int(time())

  @abstractmethod
  async def fetch(self, key: str) -> str:
    '''
    Fetches the value from cache  

    ---
    :param:  key  `str` identifies key-value pair
    '''
    raise NotImplementedError

  @abstractmethod
  async def fetch_with_ttl(self, key: str) -> Tuple[int, str]:
    '''
    Fetches the value from cache as well as remaining time to live.

    ---
    :param:  key  `str` identifies key-value pair  
    :returns:  `Tuple[int, str]`  containing timetolive value (ttl) and value
    '''
    raise NotImplementedError

  @abstractmethod
  async def put(self, key: str, value: str, ttl: Optional[int] = None):
    '''
    Puts the value within the cache with key and assigned time-to-live value

    ---
    :param:  key  `str` identifies key-value pair  
    :param:  value  `str` value to have stored identified by key  
    :param:  ttl  `int` time before value expires within cache; default: `None`
    '''
    raise NotImplementedError

  @abstractmethod
  async def clear(self, namespace: Optional[str] = None, key: Optional[str] = None) -> int:
    '''
    Clears the cache identified by given `namespace` or `key`  

    ---
    :param:  namespace  `str` identifies namespace to have entire cache cleared; default: `None`  
    :param:  key  `str` identifies key-value pair to be cleared from cache; default: `None`  
    :returns:  `int`  amount of items cleared
    '''
    raise NotImplementedError



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  config.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
'''
Module containing `CachetteConfig` class
'''
### Standard Packages ###
from typing import Callable, List, Tuple
### Third-Party Packages ###
from pydantic import ValidationError
### Local Modules ###
from fastapi_cachette.load_config import LoadConfig

class CachetteConfig(object):
  ### Basics ###
  _backend: str        = 'inmemory'
  _ttl: int            = 60

  ### Redis ###
  _redis_url: str      = None

  ### Memcached ###
  _memcached_host: str = None

  ### AWS DynamoDB & MongoDB ###
  _table_name: str     = 'fastapi-cachette'

  ### AWS DynamoDB ###
  _region: str
  _dynamodb_url: str

  ### MongoDB ###
  _database_name: str  = 'fastapi-cachette-database'
  _mongodb_url: str

  @classmethod
  def load_config(cls, settings: Callable[..., List[Tuple]]) -> 'CachetteConfig':
    try:
      config = LoadConfig(**{key.lower(): value for key, value in settings()})
      cls._backend        = config.backend or cls._backend
      cls._ttl            = config.ttl or cls._ttl
      cls._redis_url      = config.redis_url
      cls._memcached_host = config.memcached_host
      cls._table_name     = config.table_name or cls._table_name
      cls._region         = config.region
      cls._dynamodb_url   = config.dynamodb_url
      cls._database_name  = config.database_name or cls._database_name
      cls._mongodb_url    = config.mongodb_url
    except ValidationError: raise
    except Exception:
      raise TypeError('CachetteConfig must be pydantic "BaseSettings" or list of tuples')



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  backends/mongodb.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-05 14:14
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
### Standard Packages ###
from dataclasses import dataclass
from typing import Optional, Tuple
### Third-Party Packages ###
from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorCollection, AsyncIOMotorDatabase
### Local Modules ###
from fastapi_cachette.backends import Backend

@dataclass
class MongoDBBackend(Backend):
  database_name: str
  table_name: str
  ttl: int
  url: str

  @property
  def db(self) -> AsyncIOMotorDatabase:
    return AsyncIOMotorClient(self.url)[self.database_name]

  @property
  def collection(self) -> AsyncIOMotorCollection:
    return AsyncIOMotorClient(self.url)[self.database_name][self.table_name]

  @classmethod
  async def init(
      cls, database_name: str, table_name: str, ttl: int, url: str
    ) -> 'MongoDBBackend':
    client: AsyncIOMotorClient = AsyncIOMotorClient(url)
    ### Create Collection if None existed ###
    names: list = await client[database_name].list_collection_names(filter={ 'name': table_name })
    if len(names) == 0:
      await client[database_name].create_collection(table_name)
    return cls(database_name, table_name, ttl, url)

  async def fetch(self, key: str) -> str:
    document: dict = await self.collection.find_one({ 'key': key })
    value: str = None
    if document and document.get('expires', 0) > self.now: value = document.get('value', None)
    return value

  async def fetch_with_ttl(self, key: str) -> Tuple[int, str]:
    document: dict = await self.collection.find_one({ 'key': key })
    if document:
      value: str = document.get('value', None)
      ttl: int   = document.get('expires', 0) - self.now
      if ttl < 0: return 0, None
      return ttl, value
    return -1, None

  async def put(self, key: str, value: str, ttl: Optional[int] = None):
    ttl = ttl or self.ttl
    item: dict = { 'key': key, 'value': value, 'expires': self.now + ttl }
    document: dict = await self.collection.find_one({ 'key': key })
    if document:
      await self.collection.update_one({'key': key}, {'$set': item })
    else:
      await self.collection.insert_one(item)

  async def clear(self, namespace: Optional[str] = None, key: Optional[str] = None) -> int:
    count: int = 0
    if namespace:
      raise NotImplementedError
    elif key:
      document: dict = await self.collection.find_one({ 'key': key })
      if document:
        exist: bool = document.get('expires', 0) > self.now
        result = await self.collection.delete_one({'key': key})
        count += (0, 1)[exist and result.deleted_count > 0]
    return count



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  backends/inmemory.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
### Standard Packages ###
from dataclasses import dataclass
from time import time
from typing import Dict, List, Optional, Tuple
### Local Modules ###
from fastapi_cachette.backends import Backend

@dataclass
class Value:
  data: str
  expires: int

class InMemoryBackend(Backend):
  store: Dict[str, Value] = {}

  def __init__(self, ttl: Optional[int] = None):
    self.ttl = ttl
  
  async def fetch(self, key: str) -> str:
    value: Value = self.store.get(key)
    if not value: return
    elif value.expires < self.now: del self.store[key]
    else: return value.data
  
  async def fetch_with_ttl(self, key: str) -> Tuple[int, str]:
    value: Value = self.store.get(key)
    if value.expires < self.now:
      del self.store[key]
      return (0, None)
    else:
      return (value.expires - self.now, value.data)
  
  async def put(self, key: str, value: str, ttl: int = None):
    self.store[key] = Value(value, self.now + (ttl or self.ttl))

  async def clear(self, namespace: Optional[str] = None, key: Optional[str] = None) -> int:
    count: int = 0
    if namespace:
      keys: List[str] = filter(lambda key: key.startswith(namespace), self.store.keys())
      for key in keys:
        del self.store[key]
        count += 1
    elif key and self.store.get(key):
      del self.store[key]
      count += 1
    return count



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  __init__.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
'''
FastAPI extension that provides Cache Implementation Support
'''

__version__ = '0.1.1'

from .core import Cachette

__all__ = [
  'Cachette'
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['fastapi_cachette', 'fastapi_cachette.backends']

package_data = \
{'': ['*']}

install_requires = \
['fastapi>=0,<1', 'pydantic>=1.9.0,<2.0.0']

extras_require = \
{'dynamodb': ['aiobotocore[examples,dynamodb]>=2.2.0,<3.0.0'],
 'examples': ['uvicorn[examples]==0.15.0',
              'aiomcache[examples,memcached]>=0.7.0,<0.8.0',
              'aiobotocore[examples,dynamodb]>=2.2.0,<3.0.0',
              'redis[examples,redis]>=4.2.1,<5.0.0'],
 'memcached': ['aiomcache[examples,memcached]>=0.7.0,<0.8.0'],
 'mongodb': ['motor[examples,mongodb]>=2.5.1,<3.0.0'],
 'redis': ['redis[examples,redis]>=4.2.1,<5.0.0']}

setup_kwargs = {
    'name': 'fastapi-cachette',
    'version': '0.1.1',
    'description': 'Cache Implementation Extension for FastAPI Asynchronous Web Framework',
    'long_description': '# FastAPI Cachette\n\n[![Build Status](https://travis-ci.com/aekasitt/fastapi-cachette.svg?branch=master)](https://app.travis-ci.com/github/aekasitt/fastapi-cachette)\n[![Package Vesion](https://img.shields.io/pypi/v/fastapi-cachette)](https://pypi.org/project/fastapi-cachette)\n[![Format](https://img.shields.io/pypi/format/fastapi-cachette)](https://pypi.org/project/fastapi-cachette)\n[![Python Version](https://img.shields.io/pypi/pyversions/fastapi-cachette)](https://pypi.org/project/fastapi-cachette)\n[![License](https://img.shields.io/pypi/l/fastapi-cachette)](https://pypi.org/project/fastapi-cachette)\n\n## Features\n\nCache Implementation Extension for FastAPI Asynchronous Web Framework\nMost of the Backend implementation is directly lifted from [fastapi-cache](https://github.com/long2ice/fastapi-cache) by [@long2ice](https://github.com/long2ice)\n\n## Installation\n\nThe easiest way to start working with this extension with pip\n\n```bash\npip install fastapi-cachette\n# or\npoetry add fastapi-cachette\n```\n\n## Getting Started\n\nThe following examples show you how to integrate this extension to a FastAPI App\n\n### To Be Continued\n',
    'author': 'Sitt Guruvanich',
    'author_email': 'aekazitt@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/aekasitt/fastapi-cachette',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



#!/usr/bin/env python3
# coding:utf-8
# Copyright (C) 2022 All rights reserved.
# FILENAME:  backends/dynamodb.py
# VERSION: 	 0.1.1
# CREATED: 	 2022-04-03 15:31
# AUTHOR: 	 Sitt Guruvanich <aekazitt@gmail.com>
# DESCRIPTION:
#
# HISTORY:
#*************************************************************
### Standard Packages ###
from dataclasses import dataclass
from typing import Optional, Tuple
### Third-Party Packages ###
from aiobotocore.session import get_session, ClientCreatorContext
### Local Modules ###
from fastapi_cachette.backends import Backend

@dataclass
class DynamoDBBackend(Backend):
  region: str
  table_name: str
  ttl: int
  url: str

  @classmethod
  async def init(cls,
    table_name: str, ttl: int,                              \
    region: Optional[str] = None, url: Optional[str] = None \
  ) -> 'DynamoDBBackend':
    dynamodb: ClientCreatorContext = get_session() \
      .create_client('dynamodb', region_name=region, endpoint_url=url)
    async with dynamodb as client:
      ### Create Table if None Existed ###
      table_names: list = (await client.list_tables()).get('TableNames', [])
      if table_name not in table_names:
        table_definition: dict = {
          'AttributeDefinitions': [{
            'AttributeName': 'key',
            'AttributeType': 'S'
          }],
          'KeySchema': [{
            'AttributeName': 'key',
            'KeyType': 'HASH'
          }],
          'ProvisionedThroughput': {
            'ReadCapacityUnits': 10,
            'WriteCapacityUnits': 10
          }
        }
        await client.create_table(TableName=table_name, **table_definition)
    return cls(region=region, table_name=table_name, ttl=ttl, url=url)

  @property
  def dynamodb(self) -> ClientCreatorContext:
    return get_session().create_client('dynamodb', region_name=self.region, endpoint_url=self.url)

  async def fetch_with_ttl(self, key: str) -> Tuple[int, str]:
    async with self.dynamodb as client:
      response = await client.get_item(TableName=self.table_name, Key={'key': {'S': key}})
      if 'Item' in response:
        value: str = response['Item'].get('value', {}).get('S')
        ttl: int   = int(response['Item'].get('expires', {}).get('N', 0)) - self.now
        if ttl < 0:
          return 0, None
        return ttl, value

  async def fetch(self, key: str) -> str:
    async with self.dynamodb as client:
      response = await client.get_item(TableName=self.table_name, Key={'key': {'S': key}})
      if 'Item' in response:
        value: str = response['Item'].get('value', {}).get('S')
        ttl: int   = int(response['Item'].get('expires', {}).get('N', 0)) - self.now
        if ttl < 0: return None
        return value

  async def put(self, key: str, value: str, ttl: Optional[int] = None):
    ttl = ttl or self.ttl
    async with self.dynamodb as client:
      expires_at: dict = {
        'expires': { 'N': f'{ self.now + ttl }' }
      }
      await client.put_item(
        TableName=self.table_name,
        Item={ **{ 'key': { 'S': key }, 'value': { 'S': value }}, **expires_at }
      )

  async def clear(self, namespace: str = None, key: str = None) -> int:
    count: int = 0
    if namespace:
      raise NotImplementedError
    elif key:
      async with self.dynamodb as client:
        ### Checks if previously existed ###
        response: dict = await client.get_item(TableName=self.table_name, Key={'key': {'S': key}})
        ### Calculate Time-to-live ###
        ttl: int = -1
        if 'Item' in response: ttl = int(response['Item'].get('expires', {}).get('N', 0)) - self.now
        ### Sends Delete Item Request ###
        resp = await client.delete_item(TableName=self.table_name, Key={'key': {'S': key}})
        count += (0, 1)[ttl > 0 and resp['ResponseMetadata']['HTTPStatusCode'] == 200]
    return count
